#!/usr/bin/perl
#
# Calculate CO2 emissions for flights between two coordinate points on the
# earth.
# Calculate distances between two coordinate points on the earth using the
# Haversine formula.
#
# Panos Louridas, 2009--2010
#
# $Id: emcalc.pl,v 1.10 2010/08/04 08:55:38 dds Exp $
#

use strict;

use Math::Trig qw(asin);
use Getopt::Std;

use constant PI => 4 * atan2(1, 1);
use constant EARTH_RADIUS => 6372.7976;

use constant SHORT_HAUL => 3700;
use constant SHORT_HAUL_FACTOR => 0.0936;
use constant LONG_HAUL_FACTOR => 0.0819;

getopts('tan') || die "usage: $0 [-tia] (test, calc incremental, calc all)\n";
our($opt_t, $opt_a, $opt_i);

if ($opt_t) {
	my $distance = calcdist([51, 30], [0, 7],  # London
				[37, 58], [23, 43]); # Athens
	my $emissions = co2_emissions([51, 30], [0, 7],  # London
				      [37, 58], [23, 43]); # Athens
	print("London--Athens distance=${distance} km\n");
	print("London--Athens emissions=${emissions} kg\n");

	$distance = calcdist([40, 43], [-74, 0],  # NYC
			     [37, 58], [23, 43]); # Athens
	$emissions = co2_emissions([40, 43], [-74, 0],  # NYC
				   [37, 58], [23, 43]); # Athens
	print("NYC--Athens distance=${distance} km\n");
	print("NYC--Athens emissions=${emissions} kg\n");

	$distance = calcdist([3.1357, 0], [101.688, 0], # KL
			     [37, 58], [23, 43]); # Athens
	$emissions = co2_emissions([3.1357, 0], [101.688, 0], # KL
				   [37, 58], [23, 43]); # Athens
	print("KL--Athens distance=${distance} km\n");
	print("KL--Athens emissions=${emissions} kg\n");

	$distance = calcdist([-33.51, 35, 9], [151, 12, 40], # Sydney
			     [37, 58], [23, 43]); # Athens
	$emissions = co2_emissions([-33.51, 35, 9], [151, 12, 40], # Sydney
				   [37, 58], [23, 43]); # Athens
	print("Sydney--Athens distance=${distance} km\n");
	print("Sydney--Athens emissions=${emissions} kg\n");

}

if ($opt_a || $opt_i) {
	my $data_dir = 'd:/geonames';
	use DBI;
	my $dbh = DBI->connect("dbi:SQLite:dbname=$data_dir/names.sqlite");
	my $q = $dbh->prepare(q{
		select Papers.id as pid, cw.lat as clat, cw.long as clong,
		cn.name as cname, aw.lat as alat, aw.long as along,
		an.name as aname from Papers
		inner join Locations as cw on Papers.confLocId = cw.id
		inner join Locations as aw on Papers.corrLocId = aw.id
		inner join Names as cn on cw.nameId = cn.id
		inner join Names as an on aw.nameId = an.id
	} .  ($opt_i ? ' where Papers.CO2 isnull' : ''));
	$q->execute || die;
	my $u = $dbh->prepare(q{
	  UPDATE Papers SET CO2 = ? WHERE id = ?});
	my @updates;
	while (my $r = $q->fetchrow_hashref) {
		my $emissions = co2_emissions([$r->{'alat'}], [$r->{'along'}], 
					      [$r->{'clat'}], [$r->{'clong'}]);
		push(@updates, {
			'co2' => 2 * $emissions,
			'id' => $r->{'pid'}});
		print STDERR "$r->{'cname'} to $r->{'aname'} = $emissions\n";
	}
	$dbh->do('BEGIN TRANSACTION');
	for my $urow (@updates) {
		$u->execute($urow->{'co2'}, $urow->{'id'}) or die;
	}
	$dbh->do('END TRANSACTION');
}

sub coor2rads
{
	my ($deg, $min, $sec) = @_;

	return ($deg + $min / 60 + $sec / 3600) * PI / 180;
}

# Calculate the Haversine distance between two coordinate points. Code
# from
# http://www.codecodex.com/wiki/Calculate_Distance_Between_Two_Points_on_a_Globe#Perl
sub calcdist
{
	my ($lat1, $lon1, $lat2, $lon2) = @_;
	
	# $rlat1 and $rlon1 are the coordinates of the first point in radians
	# $rlat2 and $rlon2 are the coordinates of the second point in radians

	my $rlat1 = coor2rads(@$lat1);
	my $rlon1 = coor2rads(@$lon1);
	my $rlat2 = coor2rads(@$lat2);
	my $rlon2 = coor2rads(@$lon2);
		
	my $a = sin(($rlat2 - $rlat1) / 2.0);
	my $b = sin(($rlon2 - $rlon1) / 2.0);
	my $h = ($a * $a) + cos($rlat1) * cos($rlat2) * ($b * $b);
	my $theta = 2 * asin(sqrt($h)); # distance in radians
	# in order to find the distance, multiply $theta by the radius
	# of the earth, e.g.  $theta * 6,372.7976 = distance in kilometres
	# (value from http://en.wikipedia.org/wiki/Earth_radius)
	return $theta * EARTH_RADIUS;
}

# Calculate flight CO2 emissions between two earth coordinate points in radians.
# The calculation is carried out using the rules in:
# http://www.climatecrisis.net/takeaction/carboncalculator/howitwascalculated.html
sub co2_emissions
{
	my $distance = calcdist(@_);

	my $emissions = 0;
	
	if ($distance <= SHORT_HAUL) {
		$emissions = SHORT_HAUL_FACTOR * $distance;
	} else {
		$emissions = LONG_HAUL_FACTOR * $distance;
	}

	return $emissions;
}

# Local Variables: **
# mode:perl **
# perl-indent-level:8 **
# End: **
